#include "slog_pstat.h"



/*@C
    SLOG_SetPreviewStatisticsSize - Set/Override the default number of bins 
                                    of the statistics array.

  Unmodified Input parameter:
. stat_bin_size - the number of bin of the statistics array.

  Modified Input parameter :
. slog - pointer to the SLOG_STREAM where interval records are stored.

  Modified Output parameter :
. returned value - integer return status code.

  Include file need:
  slog_pstat.h

.N SLOG_RETURN_STATUS
@*/
int SLOG_SetPreviewStatisticsSize(       SLOG_STREAM      *slog,
                                   const SLOG_uint32      stat_bin_size )
{
    if ( slog == NULL ) {
        fprintf( errfile, __FILE__":SLOG_SetPreviewStatisticsSize() - \n"
                          "\t""the input SLOG_STREAM is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_SetPreviewStatisticsSize() - \n"
                          "\t""the SLOG_STREAM's Preview Statistics pointer "
                          "is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( stat_bin_size <= SLOG_PSTAT_STATSIZE_MAX )
        slog->pstat->Nbin = stat_bin_size;
    else {
        slog->pstat->Nbin = SLOG_PSTAT_STATSIZE_MAX;
        fprintf( errfile, __FILE__":SLOG_SetPreviewStatisticsSize() - \n"
                          "\t""The input number of bins, "fmt_ui32", is "
                          "too big.  Instead the maximum, "fmt_ui32", is "
                          "used\n", stat_bin_size, SLOG_PSTAT_STATSIZE_MAX );
        fflush( errfile );
    }

    return SLOG_SUCCESS;
}



SLOG_pstat_t *SLOG_PSTAT_Create( const SLOG_uint32 stat_bin_size )
{
    SLOG_pstat_t   *slog_pstat;
    int             ii;

    slog_pstat = ( SLOG_pstat_t * ) malloc( sizeof( SLOG_pstat_t ) );
    if ( slog_pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Create() - malloc() fails\n" );
        fflush( errfile );
        return NULL;
    }

    slog_pstat->Nbin       = stat_bin_size;
    slog_pstat->event_bins = ( SLOG_uint64 * ) malloc( slog_pstat->Nbin
                                                     * sizeof( SLOG_uint64 ) );
    if ( slog_pstat->Nbin > 0 && slog_pstat->event_bins == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Create() - \n"
                          "\t""malloc() fails for slog_pstat->event_bins[]\n" );
        fflush( errfile );
        return NULL;
    }

    /*  Initailize the event_bins[] array to zeros  */
    slog_pstat->bin_width_in_time = 0.0;
    for ( ii = 0; ii < slog_pstat->Nbin; ii++ )
        slog_pstat->event_bins[ ii ] = 0;

    return slog_pstat;
}



void SLOG_PSTAT_Free( SLOG_pstat_t  *slog_pstat )
{
    if ( slog_pstat != NULL ) {
        if ( slog_pstat->Nbin > 0 && slog_pstat->event_bins != NULL ) {
            free( slog_pstat->event_bins );
            slog_pstat->event_bins = NULL;
        }
        free( slog_pstat );
        slog_pstat = NULL;
    }
}



int SLOG_PSTAT_Open( SLOG_STREAM  *slog )
{
    int           ierr;

    if ( slog == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Open() - "
                          "the input SLOG_STREAM is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->hdr == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Open() - "
                          "the input SLOG_hdr_t is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    slog->pstat = SLOG_PSTAT_Create( SLOG_PSTAT_STATSIZE_MAX );
    if ( slog->pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Open() - "
                          "SLOG_PSTAT_Create() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Update the content of SLOG_hdr_t to both memory and disk  */
    slog->hdr->fptr2statistics = ( SLOG_fptr ) ftell( slog->fd );
    ierr = SLOG_WriteUpdatedHeader( slog );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Open() - "
                          "SLOG_WriteUpdatedHeader() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Write the content of SLOG_pstat_t to the disk  */
    ierr = SLOG_PSTAT_Write( slog );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Open() - "
                          "SLOG_PSTAT_Write() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    return SLOG_SUCCESS;
}



/*  The following Write() routine can be called multiple times  */
/*  But be sure the slog->hdr->fptr2statistics is set  */
int SLOG_PSTAT_Write( SLOG_STREAM  *slog )
{
    filebuf_t    *tmp_fbuf;
    SLOG_uint32   expected_bsize, bsize;
    int           ierr;
    int           ii;

    if ( slog == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - "
                          "the input SLOG_STREAM is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->hdr == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - "
                          "the input SLOG_hdr_t is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - "
                          "the input SLOG_pstat_t is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->pstat->Nbin > 0 && slog->pstat->event_bins == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - \n"
                          "\t""SLOG_pstat_t inconsistency detected\n"
                          "\t""slog->pstat->Nbin = "fmt_ui32" and "
                          "slog->pstat->event_bins[] = NULL\n",
                          slog->pstat->Nbin );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->hdr->fptr2statistics == SLOG_fptr_NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - \n"
                          "\t""the SLOG_STREAM's fptr2statistics points at "
                          "SLOG_fptr_NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Compute the expected size of the file buffer  */
    expected_bsize = SLOG_typesz[ ui32 ]
                   + slog->pstat->Nbin * SLOG_typesz[ ui64 ];

    /*  Create and initial the file buffer  */
    tmp_fbuf = ( filebuf_t * ) fbuf_create( expected_bsize );
    if ( tmp_fbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - "
                          "fbuf_create() fails!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    fbuf_filedesc( tmp_fbuf, slog->fd );

    /*  Deposit the content of SLOG_pstat_t to the file buffer  */
    bsize = fbuf_deposit( &( slog->pstat->Nbin ),
                           SLOG_typesz[ ui32 ], 1, tmp_fbuf )
          * SLOG_typesz[ ui32 ];
    for ( ii = 0; ii < slog->pstat->Nbin; ii++ )
        bsize += fbuf_deposit( &( slog->pstat->event_bins[ ii ] ),
                               SLOG_typesz[ ui64 ], 1, tmp_fbuf )
               * SLOG_typesz[ ui64 ];

    /*  File Buffer usage test  */
    if ( bsize != expected_bsize ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - "
                          "bsize written("fmt_ui32") != expected_bsize "
                          "written("fmt_ui32")\n", bsize, expected_bsize );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Empty the content of tmp_fbuf to the disk at the specified location  */
    ierr = fseek( slog->fd, slog->hdr->fptr2statistics, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Write() - "
                          "fseek( slog->hdr->fptr2statistics ) fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    fbuf_empty( tmp_fbuf );
    fflush( slog->fd );

    fbuf_free( tmp_fbuf );

    return SLOG_SUCCESS;
}



int SLOG_PSTAT_Read( SLOG_STREAM  *slog )
{
    SLOG_fptr      file_loc_saved;
    filebuf_t     *tmp_fbuf;
    SLOG_uint32    expected_bsize, bsize;
    SLOG_uint32    pstat_Nbin;
    int            ierr;
    int            ii;

    if ( slog == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - "
                          "the input SLOG_STREAM is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->hdr == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - "
                          "the input SLOG_hdr_t is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->hdr->fptr2statistics == SLOG_fptr_NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - \n"
                          "\t""the SLOG_STREAM's fptr2statistics points at "
                          "SLOG_fptr_NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Save the current file pointer position  */
    file_loc_saved = ( SLOG_fptr ) ftell( slog->fd );

    /*  Locate where to read the Preview Statistics Array  */
    ierr = fseek( slog->fd, slog->hdr->fptr2statistics, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - "
                          "fseek( slog->hdr->fptr2statistics ) fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Read in the size of the Preview Statistics Array  */
    ierr = bswp_fread( &( pstat_Nbin ),
                       sizeof( SLOG_uint32 ), 1, slog->fd );
    if ( ierr < 1 ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - \n"
                          "\t""Fails at reading the number of bins "
                          "of the Preview Statistics Array\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Compute the expected size of the file buffer  */
    expected_bsize = pstat_Nbin * SLOG_typesz[ ui64 ];

    /*  Create and Initialize the file buffer  */
    tmp_fbuf = ( filebuf_t * ) fbuf_create( expected_bsize );
    if ( tmp_fbuf == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - "
                          "fbuf_create() fails!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    fbuf_filedesc( tmp_fbuf, slog->fd );

    /*  Fill up the file buffer from the disk  */
    if ( fbuf_refill( tmp_fbuf ) == EOF ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - \n"
                          "\t""fbuf_refill( tmp_fbuf ) fails\n" );
        fflush( errfile );
        return SLOG_EOF;
    }

    /*  Create the SLOG_pstat_t data structure in SLOG_STREAM  */
    slog->pstat = SLOG_PSTAT_Create( pstat_Nbin );
    if ( slog->pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Init() - "
                          "SLOG_PSTAT_Create() fails\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Withdraw SLOG_pstat_t from the tmp_fbuf  */
    bsize = 0;
    for ( ii = 0; ii < slog->pstat->Nbin; ii++ )
        bsize += fbuf_withdraw( &( slog->pstat->event_bins[ ii ] ),
                                SLOG_typesz[ ui64 ], 1, tmp_fbuf )
               * SLOG_typesz[ ui64 ];

    /*  File Buffer usage test  */
    if ( bsize != expected_bsize ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - "
                          "bsize written("fmt_ui32") != expected_bsize "
                          "written("fmt_ui32")\n", bsize, expected_bsize );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*  Restore the saved file pointer position  */
    ierr = fseek( slog->fd, file_loc_saved, SEEK_SET );
    if ( ierr != 0 ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Read() - fseek() "
                          "fails, cannot restore the orginal file position "
                          "before the end of the subroutine call\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }
    fbuf_free( tmp_fbuf );

    return SLOG_SUCCESS;
}



int SLOG_PSTAT_SetBinWidthInTime( SLOG_STREAM  *slog )
{
    SLOG_pstat_t  *pstat;
    SLOG_dir_t    *dir;

    if ( slog == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_SetBinWidthInTime() - "
                          "the input SLOG_STREAM is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_SetBinWidthInTime() - "
                          "the input SLOG_pstat_t is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if (    ( slog->frame_dir ).Nframe > 0 
         && ( slog->frame_dir ).entries == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_SetBinWidthInTime() - "
                          "the input SLOG_dir_t is inconsistent!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( ( slog->frame_dir ).Nframe < 1 )
        return SLOG_SUCCESS;

    dir   = &( slog->frame_dir );
    pstat = slog->pstat;

    if ( pstat->Nbin > 0 )
        pstat->bin_width_in_time = ( ( dir->entries[ dir->Nframe-1 ] ).endtime
                                   - ( dir->entries[ 0 ] ).starttime )
                                 / pstat->Nbin;
    else
        pstat->bin_width_in_time = ( ( dir->entries[ dir->Nframe-1 ] ).endtime
                                   - ( dir->entries[ 0 ] ).starttime );

    return SLOG_SUCCESS;
}

#define COUNT_EVENT_STARTTIME 1

int SLOG_PSTAT_Update(       SLOG_pstat_t           *pstat,
                       const SLOG_intvlrec_blist_t  *src_bbuf )
{
    const SLOG_intvlrec_lptr_t  *lptr;
          SLOG_time              irec_endtime;
          SLOG_uint32            bin_idx;
          SLOG_time              bin_width_in_time;

    if ( pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Update() - "
                          "the input SLOG_pstat_t pointer is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    bin_width_in_time = pstat->bin_width_in_time;

    if ( bin_width_in_time <= 0.0 ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Update() - \n"
                          "\t""the input SLOG_pstat_t's bin_width_in_time = "
                          fmt_time"\n", bin_width_in_time );
        fflush( errfile );
        return SLOG_FAIL;
    }

    for ( lptr = src_bbuf->lhead; lptr != NULL; lptr = lptr->next ) {
#if defined( COUNT_EVENT_STARTTIME )
         bin_idx = lptr->irec->starttime / bin_width_in_time;
#else
         irec_endtime = lptr->irec->starttime + lptr->irec->duration;
         bin_idx = irec_endtime / bin_width_in_time;
#endif
         pstat->event_bins[ bin_idx ]++;
    }

    return SLOG_SUCCESS;
}



int SLOG_PSTAT_Close( SLOG_STREAM  *slog )
{
    int    ierr;

    if ( slog == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Close() - "
                          "the input SLOG_STREAM is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    if ( slog->pstat == NULL ) {
        fprintf( errfile, __FILE__":SLOG_PSTAT_Close() - "
                          "the input SLOG_pstat_t is NULL\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    /*
        Extra Increment statement to guarantee the sum of real records
        count is the same as the sum of real records given by the user.
    */
#if ! defined( COUNT_EVENT_STARTTIME )
    slog->pstat->event_bins[ slog->pstat->Nbin-1 ]++;
#endif

    /*  Write the updated statistics info onto the disk again  */
    ierr = SLOG_PSTAT_Write( slog );
    if ( ierr != SLOG_SUCCESS ) {
        fprintf( errfile, __FILE__":SLOG_CloseOutputStream() - "
                          "SLOG_PSTAT_Write( slog ) fails!!\n" );
        fflush( errfile );
        return SLOG_FAIL;
    }

    return SLOG_SUCCESS;
}



void SLOG_PSTAT_Print( FILE *outfd, const SLOG_pstat_t  *pstat )
{
    const   int   Nitems_at_line = 5;
            int   ii;

    for ( ii = 0; ii < pstat->Nbin; ii++ ) {
        fprintf( outfd, fmt_ui64", ", pstat->event_bins[ ii ] );
        if ( ii % Nitems_at_line == Nitems_at_line - 1 )
            fprintf( outfd, "\n" );
    }
    fprintf( outfd, "\n" );
    fflush( outfd );
}
